﻿using System;
using System.Collections.Generic;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using Clarion.Framework.Extensions;
using WorldServerLibrary.Model;
using System.Threading;
using System.Diagnostics;

namespace ClarionSimulation
{
    /// <summary>

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction, Thing whichThing);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {
        DO_NOTHING,
        ROTATE_CLOCKWISE,
		EAT_FOOD,
		COLLECT_JEWEL,
		GO_POSITION,
		DELIVER_LEAFLET,
		CONTOUR_OBSTACLE,
		CREATE_THINGS
    }
	
	
	public enum jewelColors
	{
		red,
		green,
		blue,
		yellow,
		magenta,
		white,
		NUM_OF_COLORS
	}
	
	
    public class ClarionAgent
    {
        #region Constants

        /// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// <summary>
        /// Constant that represents that there is at least one wall ahead
        /// </summary>
        private String DIMENSION_WALL_AHEAD = "WallAhead";
		
		private double MAXIMUM_DISTANCE = 1000.0;
		private double COLLECT_DISTANCE = 30.0;
		private double CONTOUR_DISTANCE = 30.0;
		private double DELIVER_DISTANCE = 61.0;
		
		private double MINIMUM_FUEL = 300.0;
		
		private String EAT_FOOD = "eatFood";
		private String FULFILL_LEAFLET = "fulfillLeaflet";

        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;
		
        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;
		
		
		private Creature currentCreature;
		
		private GoalChunk eatFood;
		private GoalChunk fulfillLeaflet;
		
		private SimplifiedQBPNetwork movementNet;
		
		private Thing deliverySpot = null;
		
		private Leaflet completedLeaflet = null;
		
		
        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a wall ahead
        /// </summary>
        private DimensionValuePair InputWallAhead;
		
		private DimensionValuePair leafletCompletion;
		
		private DimensionValuePair pos_x;
		private DimensionValuePair pos_y;
		
		private DimensionValuePair missing_red;
		private DimensionValuePair missing_green;
		private DimensionValuePair missing_blue;
		private DimensionValuePair missing_yellow;
		private DimensionValuePair missing_magenta;
		private DimensionValuePair missing_white;
		
		

		private List<Thing> jewels;
		private List<Thing> foods;
		
		private List<Leaflet> leaflets;
		
		private List<Thing> collectedJewels;
		
		private int[] jewelsNeeded;
		private double[] jewelsActivation;
		
		
		
        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
        private ExternalActionChunk OutputRotateClockwise;

        private ExternalActionChunk outputDeliverLeaflet;
		
		private ExternalActionChunk outputContourObstacle;

        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {
			World.LoggingLevel = TraceLevel.Verbose;
			
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            InputWallAhead = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);
			
			leafletCompletion = World.NewDimensionValuePair ("leaflet", "completion");
			
			pos_x = World.NewDimensionValuePair ("creature", "x");
			pos_y = World.NewDimensionValuePair ("creature", "y");
			
			missing_red     = World.NewDimensionValuePair ("missing", "red");
		    missing_green   = World.NewDimensionValuePair ("missing", "green");
			missing_blue    = World.NewDimensionValuePair ("missing", "blue");
			missing_yellow  = World.NewDimensionValuePair ("missing", "yellow");
			missing_magenta = World.NewDimensionValuePair ("missing", "magenta");
			missing_white   = World.NewDimensionValuePair ("missing", "white");
				

            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk (ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
			OutputRotateClockwise.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.ROTATE_CLOCKWISE.ToString()));

            outputDeliverLeaflet = World.NewExternalActionChunk (ClarionAgentActionType.DELIVER_LEAFLET.ToString());
			outputDeliverLeaflet.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.DELIVER_LEAFLET.ToString()));
			
			outputContourObstacle = World.NewExternalActionChunk (ClarionAgentActionType.CONTOUR_OBSTACLE.ToString());
			outputContourObstacle.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.CONTOUR_OBSTACLE.ToString()));
			
            //Create thread to simulation
            runThread = new Thread (RunThread);
			Console.WriteLine ("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {	
			jewels = new List<Thing>();
			foods = new List<Thing>();
			
			collectedJewels = new List<Thing>();

			jewelsNeeded     = new int[(int) jewelColors.NUM_OF_COLORS];
			jewelsActivation = new double[(int) jewelColors.NUM_OF_COLORS];
			
            // Setup the ACS Subsystem
            SetupACS();                    
        }

        private void SetupMS()
        {            
            //RichDrive
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
        {

            // Create contour obstacle rule

			SupportCalculator contourObstacleSupportCalculator = FixedRuleDelegateToContourObstacle;
			
            FixedRule ruleContourObstacle = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, outputContourObstacle, contourObstacleSupportCalculator);
			
			/*
			ruleAvoidCollisionWall.Parameters.PARTIAL_MATCH_ON = true;
			ruleAvoidCollisionWall.Parameters.DEFAULT_PARTIAL_MATCH_ON_THRESHOLD = 0.1;
			ruleAvoidCollisionWall.Parameters.PARTIAL_MATCH_THRESHOLD = 0.1;
			ruleAvoidCollisionWall.Parameters.POSITIVE_MATCH_THRESHOLD = 0.1;
			*/
			
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit (ruleContourObstacle);

            // Create deliver leaflet rule
			
            SupportCalculator deliverLeafletSupportCalculator = fixedRuleDelegateToDeliverLeaflet;
            FixedRule ruleDeliverLeaflet = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, outputDeliverLeaflet, deliverLeafletSupportCalculator);		
			
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit (ruleDeliverLeaflet);
			
			
			
            // Disable Rule Refinement
            // CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selection type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE  = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE  = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
			
			// Create goals for leaflet fetching process
			
			eatFood = World.NewGoalChunk(EAT_FOOD);			
			fulfillLeaflet = World.NewGoalChunk(FULFILL_LEAFLET);

			// Create the action chunks for setting each of those goals
			
            GoalStructureUpdateActionChunk setEatFood = World.NewGoalStructureUpdateActionChunk();
            GoalStructureUpdateActionChunk setFulfillLeaflet = World.NewGoalStructureUpdateActionChunk();

            setEatFood.Add (GoalStructure.RecognizedActions.SET_RESET, eatFood);
            setFulfillLeaflet.Add (GoalStructure.RecognizedActions.SET_RESET, fulfillLeaflet);
			
			// Create the drives
			
			DeficitChangeProcessor eatFoodAndFulfillLeafletDrivesDeficitChangeProcessor = EatFoodAndFulFillLeafletDrives_DeficitChange;

			// Eat Food drive
			
            FoodDrive eatFoodDrive = AgentInitializer.InitializeDrive (CurrentAgent, 
			                                                           FoodDrive.Factory, 
			                                                           .5, 
			                                                           eatFoodAndFulfillLeafletDrivesDeficitChangeProcessor);

            DriveEquation eatFoodDrive_equation = AgentInitializer.InitializeDriveComponent (eatFoodDrive, DriveEquation.Factory);

            eatFoodDrive.Commit (eatFoodDrive_equation);
			
			
			// Fulfill leaflet drive
			
            RecognitionAchievementDrive fulfillLeafletDrive = AgentInitializer.InitializeDrive (CurrentAgent,
			                                                                                    RecognitionAchievementDrive.Factory, 
			                                                                                    .5, 
			                                                                                    eatFoodAndFulfillLeafletDrivesDeficitChangeProcessor);

            DriveEquation fulfillLeafletDrive_equation = AgentInitializer.InitializeDriveComponent (fulfillLeafletDrive, DriveEquation.Factory);

            fulfillLeafletDrive.Commit(fulfillLeafletDrive_equation);
			
						
            CurrentAgent.Commit (eatFoodDrive);
			CurrentAgent.Commit (fulfillLeafletDrive);
			
			
			// Initialize the meta-cognitive module
			
			
			GoalSelectionModule gsm = AgentInitializer.InitializeMetaCognitiveModule (CurrentAgent, GoalSelectionModule.Factory);

            GoalSelectionEquation gse = AgentInitializer.InitializeMetaCognitiveDecisionNetwork (gsm, GoalSelectionEquation.Factory);

            gse.Input.Add (eatFoodDrive.GetDriveStrength());
            gse.Input.Add (fulfillLeafletDrive.GetDriveStrength());

            gse.Output.Add (setEatFood);
            gse.Output.Add (setFulfillLeaflet);

            gse.SetRelevance (setEatFood, eatFoodDrive, 1);	
            gse.SetRelevance (setFulfillLeaflet, fulfillLeafletDrive, 1);

            gsm.Commit (gse);

            CurrentAgent.Commit (gsm);

			CurrentAgent.MS.Parameters.CURRENT_GOAL_ACTIVATION_OPTION = MotivationalSubsystem.CurrentGoalActivationOptions.FULL;
			
			
			// Setup the movement network
			
			movementNet = AgentInitializer.InitializeImplicitDecisionNetwork (CurrentAgent, SimplifiedQBPNetwork.Factory);
			

			// Add the goals as input for the movement network
			
			movementNet.Input.Add (eatFood);
			movementNet.Input.Add (fulfillLeaflet);
			
			movementNet.Input.Add (InputWallAhead);
			movementNet.Input.Add (leafletCompletion);
			
			
			// Add the elements representing the current creature position
			
			movementNet.Input.Add (pos_x);
			movementNet.Input.Add (pos_y);
			
			
			// Add the elements representing the missing jewels in the existing leaflets
			
			movementNet.Input.Add (missing_red);
			movementNet.Input.Add (missing_green);
			movementNet.Input.Add (missing_blue);
			movementNet.Input.Add (missing_yellow);
			movementNet.Input.Add (missing_magenta);
			movementNet.Input.Add (missing_white);
			
			
			// Add the rule for rotating clockwise
			
			movementNet.Output.Add (OutputRotateClockwise);
			
			movementNet.Parameters.LEARNING_RATE = 1;
			
			CurrentAgent.Commit (movementNet);
			
            RefineableActionRule.GlobalParameters.SPECIALIZATION_THRESHOLD_1 = -.6;
            RefineableActionRule.GlobalParameters.GENERALIZATION_THRESHOLD_1 = -.1;
            RefineableActionRule.GlobalParameters.INFORMATION_GAIN_OPTION = RefineableActionRule.IGOptions.PERFECT;

        }
		
		
        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        /// 
        
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation)
        {
			// Check if hasn't created yet the rule for going until the delivery spot
			
			if (deliverySpot == null)
			{
				deliverySpot = sensorialInformation.First (item => (item.CategoryId == Thing.CATEGORY_DeliverySPOT));
				
				
				// Create an external action to go until the delivery spot
				
				ExternalActionChunk outputGoDeliverySpot = World.NewExternalActionChunk ("go_deliverySpot");
				
				outputGoDeliverySpot.Add (World.NewDimensionValuePair ("x", deliverySpot.X1));
				outputGoDeliverySpot.Add (World.NewDimensionValuePair ("y", deliverySpot.Y1));
				outputGoDeliverySpot.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.GO_POSITION.ToString()));
				
				// Update the movement network
			
				CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
				
				movementNet.Output.Add (outputGoDeliverySpot);
				
				CurrentAgent.Commit (movementNet);
				
				Console.WriteLine (">>>> Added delivery spot ({0}, {1}) into the agent", deliverySpot.X1, deliverySpot.Y1);
			}
			
			
			
			// Save the current creature state information
			currentCreature = (Creature) sensorialInformation.First (item => (item is Creature));

            // New sensory information
            SensoryInformation si = World.NewSensoryInformation(CurrentAgent);

            // Detect if we have a wall ahead
            Boolean wallAhead = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= CONTOUR_DISTANCE)).Any();
            
			double wallAheadActivationValue = wallAhead ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
            
			
			// Stores the new jewels of foods for further reference
			
			foreach (Thing visualizedItem in sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_FOOD) ||
			                                                                      (item.CategoryId == Thing.CATEGORY_NPFOOD) ||
			                                                                      (item.CategoryId == Thing.CATEGORY_PFOOD))))
			{
				Thing foundFood = foods.Find (item => item.Name == visualizedItem.Name);
				
				if (foundFood == null)
				{
					foods.Add (visualizedItem);
					
					CreateRulesForFood (visualizedItem);
				}
				else
				{
					// Check if food has been moved
					
					if ((foundFood.X1 != visualizedItem.X1) || (foundFood.Y1 != visualizedItem.Y1))
					{
						// Remove all the info related to the old food position...
						
						removeCollectedOrEaten (visualizedItem.Name);
						
						Thread.Sleep (200);
						
						
						// Now add info related to the new position
						
						foods.Add (visualizedItem);
						
						CreateRulesForFood (visualizedItem);
					}
				}
			}
			

			foreach (Thing visualizedItem in sensorialInformation.Where (item => item.CategoryId == Thing.CATEGORY_JEWEL))
			{
				Thing foundJewel = jewels.Find (item => item.Name == visualizedItem.Name);
				
				if (foundJewel == null)
				{
					jewels.Add (visualizedItem);
					
					CreateRulesForJewel (visualizedItem);					
				}
				else
				{
					// Check if jewel has been moved
					
					if ((foundJewel.X1 != visualizedItem.X1) || (foundJewel.Y1 != visualizedItem.Y1))
					{
						// Remove all the info related to the old jewel position...
						
						removeCollectedOrEaten (visualizedItem.Name);
						
						Thread.Sleep (200);
						
						
						// Now add info related to the new position
						
						jewels.Add (visualizedItem);
						
						CreateRulesForJewel (visualizedItem);
					}
					else
					{						
						// Check if jewel has changed color
						
						if (foundJewel.Material.Color != visualizedItem.Material.Color)
						{
							// Update it
							
							foundJewel.Material.Color = visualizedItem.Material.Color;
						}
					}
				}
			}
			
			
			// Stores the received leaflets; updates the list just to make sure the most recent ones are stored
			
			leaflets = new List<Leaflet>();
			
			foreach (Thing visualizedItem in sensorialInformation.Where (item => item.CategoryId == Thing.CATEGORY_LEAFLET))
			{
				Leaflet newLeaflet = (Leaflet) visualizedItem;
				
				if (leaflets.Find (item => item.ID == newLeaflet.ID) == null)
				{
					leaflets.Add (newLeaflet);
				}
			}
			
			
			
			// Updates the leaflets counters
			
			completedLeaflet = CheckCompleteLeaflet ();
			
			
			si.Add (InputWallAhead, wallAheadActivationValue);
			
			si.Add (leafletCompletion, completedLeaflet == null ? 0.0 : 1.0);
			
			si.Add (pos_x, currentCreature.X1 / 800);
			si.Add (pos_y, currentCreature.Y1 / 600);
			        
			si.Add (missing_red, jewelsActivation[(int) jewelColors.red]);
			si.Add (missing_green, jewelsActivation[(int) jewelColors.green]);
			si.Add (missing_blue, jewelsActivation[(int) jewelColors.blue]);
			si.Add (missing_yellow, jewelsActivation[(int) jewelColors.yellow]);
			si.Add (missing_magenta, jewelsActivation[(int) jewelColors.magenta]);
			si.Add (missing_white, jewelsActivation[(int) jewelColors.white]);
				
			
			// Sets the drive stimulus, based on the current creature fuel level

			si[typeof(FoodDrive), FoodDrive.MetaInfoReservations.STIMULUS] = (MINIMUM_FUEL / currentCreature.Fuel) < 1 ? 0 : 1;
			si[typeof(RecognitionAchievementDrive), RecognitionAchievementDrive.MetaInfoReservations.STIMULUS] = ((currentCreature.Fuel - MINIMUM_FUEL) / MINIMUM_FUEL) > 0 ? 1 : 0;
			
            return si;
        }
		
		
		
		private Leaflet CheckCompleteLeaflet ()
		{
			Leaflet completeLeaflet = null;
			
			for (int i = 0; i < jewelsNeeded.Length; i++)
			{
				jewelsNeeded[i] = 0;
			}
			
			
			// Go through all the leaflets to find the first completed one
			
			foreach (Leaflet currentLeaflet in leaflets)
			{
				bool complete = true;
				
				foreach (LeafletItem jewelColor in currentLeaflet.items)
				{
					jewelsNeeded[(int) Enum.Parse(typeof (jewelColors), jewelColor.color, true)] += jewelColor.desiredQuantity; 
					
					if (jewelColor.desiredQuantity > collectedJewels.FindAll (item => item.Material.Color == jewelColor.color).Count)
					{
						complete = false;
					}
				}
				
				// Check if current leaflet is complete...
				
				if ((complete == true) && (completeLeaflet == null))
				{
					// ...if so, save it to return.
					
					completeLeaflet = currentLeaflet;
					
					Console.WriteLine ("Leaflet {0} is now completed !", currentLeaflet.ID);
				}
			}
			
			
			for (int i = 0; i < jewelsActivation.Length; i++)
			{
				int currentJewelsCount = collectedJewels.FindAll (item => item.Material.Color == Enum.GetName (typeof (jewelColors), i)).Count;
				
				if (jewelsNeeded[i] == 0 || jewelsNeeded[i] == currentJewelsCount)
				{                                               
					jewelsActivation[i] = 0.0;
				}
				else
				{
					jewelsActivation[i] = 1.0 - (currentJewelsCount + 0.1 / jewelsNeeded[i]);
				}
			}
			
			return completeLeaflet;
		}
		
		
		
		private void CreateRulesForFood (Thing whichFood)
		{
			// Common information
			
			DimensionValuePair foodName = World.NewDimensionValuePair ("food", whichFood.Name);
			DimensionValuePair foodX    = World.NewDimensionValuePair ("x", whichFood.X1);
			DimensionValuePair foodY    = World.NewDimensionValuePair ("y", whichFood.Y1);

			
			// External action for going until this food
			
			ExternalActionChunk outputGoFood = World.NewExternalActionChunk ("go_" + whichFood.Name);
			
			outputGoFood.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.GO_POSITION.ToString()));
			outputGoFood.Add (foodName);
			outputGoFood.Add (foodX);
			outputGoFood.Add (foodY);

			
			// Fixed Rule for eating this food
			
			ExternalActionChunk outputEatFood = World.NewExternalActionChunk ("eat_" + whichFood.Name);
			
			outputEatFood.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.EAT_FOOD.ToString()));
			outputEatFood.Add (foodName);
			outputEatFood.Add (foodX);
			outputEatFood.Add (foodY);

			SupportCalculator collectJewelSupportCalculator = FixedRuleDelegateToCollectOrEat;
			
            FixedRule ruleEatJewel = AgentInitializer.InitializeActionRule (CurrentAgent, 
			                                                                FixedRule.Factory, 
			                                                                outputEatFood, 
			                                                                collectJewelSupportCalculator);
			                                                                    			
			
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit (ruleEatJewel);
			
			
			// Update the movement network
			
			CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
			
			movementNet.Output.Add (outputGoFood);
			
			CurrentAgent.Commit (movementNet);
			
			Console.WriteLine (">>>> Added food {0} into the agent", whichFood.Name);

		}
		
		
		
		private void CreateRulesForJewel (Thing whichJewel)
		{
			// Common information
			
			DimensionValuePair jewelName = World.NewDimensionValuePair ("jewel", whichJewel.Name);
			DimensionValuePair jewelX    = World.NewDimensionValuePair ("x", whichJewel.X1);
			DimensionValuePair jewelY    = World.NewDimensionValuePair ("y", whichJewel.Y1);
			
			
			// External action for going until this jewel
			
			
			ExternalActionChunk outputGoJewel = (ExternalActionChunk) World.NewExternalActionChunk ("go_" + whichJewel.Name);
			
			
			outputGoJewel.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.GO_POSITION.ToString()));
			outputGoJewel.Add (jewelName);
            outputGoJewel.Add (World.NewDimensionValuePair ("color", whichJewel.Material.Color));
			outputGoJewel.Add (jewelX);
			outputGoJewel.Add (jewelY);
			

			// Fixed Rule for collecting this jewel
			
			ExternalActionChunk outputCollectJewel = World.NewExternalActionChunk ("collect_" + whichJewel.Name);
			
			outputCollectJewel.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.COLLECT_JEWEL.ToString()));
			outputCollectJewel.Add (jewelName);
			outputCollectJewel.Add (jewelX);
			outputCollectJewel.Add (jewelY);

			SupportCalculator eatFoddSupportCalculator = FixedRuleDelegateToCollectOrEat;
			
            FixedRule ruleCollectJewel = AgentInitializer.InitializeActionRule (CurrentAgent, 
			                                                                    FixedRule.Factory, 
			                                                                    outputCollectJewel, 
			                                                                    eatFoddSupportCalculator);
			                                                                    			
			
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleCollectJewel);
			
			
			// Update the movement network
			
			CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
			
			movementNet.Output.Add (outputGoJewel);
			
			CurrentAgent.Commit (movementNet);
			
			Console.WriteLine (">>>> Added jewel {0} into the agent", whichJewel.Name);
		}
		
		
		
		/*
		private void removeCollectedJewel (string whichJewel)
		{	
			// Remove jewel related rules
			
			IEnumerable<ICommitable> allRules = CurrentAgent.GetInternals (Agent.InternalContainers.ACTION_RULES);
												
			foreach (ICommitable rule in allRules)
			{
				if ((rule is FixedRule) && ((FixedRule) rule).OutputChunk.LabelAsIComparable.ToString() == "collect_" + whichJewel)
				{
					CurrentAgent.Retract (rule, Agent.InternalContainers.ACTION_RULES);
					CurrentAgent.Remove (rule);
				}
			}
	
			
			// Remove output from movement network
			
			CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
			
			
			ExternalActionChunk outputGoJewel = (ExternalActionChunk) World.GetActionChunk ("go_" + whichJewel);
			
			movementNet.Output.Remove (outputGoJewel);
			
			CurrentAgent.Commit (movementNet);			
			
			
			IEnumerator<DimensionValuePair> iItems = outputGoJewel.GetEnumerator();
			
			while (iItems.MoveNext())
			{
				DimensionValuePair item = iItems as DimensionValuePair;
				
				World.Remove (item);
			}
				
			World.Remove (outputGoJewel);
			
				
			ExternalActionChunk outputCollectJewel = (ExternalActionChunk) World.GetActionChunk ("collect_" + whichJewel);
			
			World.Remove (outputCollectJewel.First (item => item.AsDimensionValuePair.Dimension.ToString() == "type"));
			
			World.Remove (outputCollectJewel);
			
			
			
			Thing whichJewelToRemove = jewels.Find (item => item.Name == whichJewel);
			
			collectedJewels.Add (whichJewelToRemove);
			
			jewels.Remove (whichJewelToRemove);
			
			
			Console.WriteLine (">>>>>>>>>>>>>>>>>>>> Remove jewel {0}", whichJewel);
		}
		*/
		

		
		
		private void removeCollectedOrEaten (string whichThing)
		{	
			string token;
			bool isJewel = false;
			
			// Remove thing related rules
			
			IEnumerable<ICommitable> allRules = CurrentAgent.GetInternals (Agent.InternalContainers.ACTION_RULES);
						
			if (whichThing.IndexOf ("Jewel") >= 0)
			{
				token   = "collect_";
				isJewel = true;
			}
			else
			{
				token = "eat_";
			}
			
			
			foreach (ICommitable rule in allRules)
			{
				if ((rule is FixedRule) && ((FixedRule) rule).OutputChunk.LabelAsIComparable.ToString() == token + whichThing)
				{
					CurrentAgent.Retract (rule, Agent.InternalContainers.ACTION_RULES);
					CurrentAgent.Remove (rule);
				}
			}
	
			
			// Remove output from movement network
			
			CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
			
			
			ExternalActionChunk outputGoThing = (ExternalActionChunk) World.GetActionChunk ("go_" + whichThing);
			
			movementNet.Output.Remove (outputGoThing);
			
			CurrentAgent.Commit (movementNet);			
			
			
			IEnumerator<DimensionValuePair> iItems = outputGoThing.GetEnumerator();
			
			while (iItems.MoveNext())
			{
				DimensionValuePair item = iItems.Current;
				
				World.Remove (item);
			}
				
			World.Remove (outputGoThing);
			
				
			ExternalActionChunk outputCollectOrEat = (ExternalActionChunk) World.GetActionChunk (token + whichThing);
			
			
			iItems = outputCollectOrEat.GetEnumerator();
			
			while (iItems.MoveNext())
			{
				DimensionValuePair item = iItems.Current;
				
				World.Remove (item);
			}
									
			World.Remove (outputCollectOrEat);
			
			
			
			if (isJewel == true)
			{
				Thing whichThingToRemove = jewels.Find (item => item.Name == whichThing);
	
				collectedJewels.Add (whichThingToRemove);
				
				jewels.Remove (whichThingToRemove);
			}
			else
			{
				Thing whichThingToRemove = foods.Find (item => item.Name == whichThing);

				foods.Remove (whichThingToRemove);
			}
			
			
			Console.WriteLine (">>>>>>>>>>>>>>>>>>>> Removed thing {0}", whichThing);
		}

		
		
		
		
        #endregion

        #region Delegate Methods
		
		#region Deficit change 
		
		// Update the drive defcit according to the current goal set
		
		private double EatFoodAndFulFillLeafletDrives_DeficitChange (ActivationCollection currentInput, Drive target)
		{
			GoalChunk currentGoal = CurrentAgent.CurrentGoal;
			
            if (currentGoal != null)
            {
				// Check if the current goal is aligned with the drive which deficit is being updated...
				
                if ((currentGoal.LabelAsIComparable.ToString() == FULFILL_LEAFLET && target is RecognitionAchievementDrive) ||
                    (currentGoal.LabelAsIComparable.ToString() == EAT_FOOD && target is FoodDrive))
				{
					
					// ...if so, decrease the deficit by a 0.1% rate
					
                    target.Parameters.DEFICIT_CHANGE_RATE = .999;
				}
                else
				{	
					// ...if not, increase the drive deficit by a 0.1% rate
					
                    target.Parameters.DEFICIT_CHANGE_RATE = 1.001;
				}
            }

            return target.Deficit * target.Parameters.DEFICIT_CHANGE_RATE;
		}
		
		#endregion
			
		
        #region Fixed Rules	
		
		
		
		private double FixedRuleDelegateToContourObstacle (ActivationCollection currentInput, Rule target)
        {
			double activation = ((currentInput.Contains (InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0;
			
            // See partial match threshold to verify what are the rules available for action selection
			
			Console.WriteLine ("Contour: {0}", activation);
			
            return activation;
        }

		
		
		
		private double fixedRuleDelegateToDeliverLeaflet (ActivationCollection currentInput, Rule target)
        {
			Double distance = Double.PositiveInfinity;
			double activation = 0.0;
			
			distance = Math.Sqrt(Math.Pow (currentCreature.X1 - deliverySpot.X1, 2) + Math.Pow (currentCreature.Y1 - deliverySpot.Y1, 2)); 
				
			activation = ((distance < DELIVER_DISTANCE) && (completedLeaflet != null)) ? 1.0 : 0.0;
			
			
			Console.WriteLine ("DELIVER LEAFLET - Activation: {0}", activation);
			
            return activation;
        }

		
		
		private double FixedRuleDelegateToCollectOrEat (ActivationCollection currentInput, Rule target)
        {
			Double distance = Double.PositiveInfinity;
			double activation = 0;
			
			ExternalActionChunk outputCollectOrEat = (ExternalActionChunk) target.OutputChunk;
			
			double thing_x = System.Convert.ToDouble (outputCollectOrEat.First (item => item.AsDimensionValuePair.Dimension.ToString() == "x").Value.AsIComparable);
			double thing_y = System.Convert.ToDouble (outputCollectOrEat.First (item => item.AsDimensionValuePair.Dimension.ToString() == "y").Value.AsIComparable);

			// Get the thing name, removing the leading "go_"...
			
			string name = outputCollectOrEat.LabelAsIComparable.ToString().Substring (outputCollectOrEat.LabelAsIComparable.ToString().IndexOf ("_") + 1);
				
			
			if (((name.IndexOf ("Jewel") >= 0) && (jewels.Find (item => item.Name == name) != null)) ||
			    ((name.IndexOf ("Food") >= 0) && (foods.Find (item => item.Name == name) != null)))
			{
				distance = Math.Sqrt(Math.Pow (thing_x - currentCreature.X1, 2) + Math.Pow (thing_y - currentCreature.Y1, 2)); 
				
				activation = distance < COLLECT_DISTANCE ? 1.0 : 0.0;
			}
			
			Console.WriteLine ("COLLECT/EAT ({0}, {1}) - thing: {2} - distance: {3} - Activation: {4}", thing_x, thing_y, name, distance, activation);
			
            return activation;
        }

		
		#endregion
		
		
        #region Run Thread Method

        private void RunThread(object obj)
        {
			ExternalActionChunk chosen = ExternalActionChunk.DO_NOTHING;
			
			double lastMovementFeedback = 0.0;
			
			bool tryingToContour = false;
			
		
			
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
			
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
				//Console.WriteLine ("+++ Cicle: {0}", CurrentCognitiveCycle);
				
				
				if (Console.KeyAvailable == true)
				{
					if (Console.ReadKey (true).Key == ConsoleKey.Spacebar)
					{
						OnNewExternalActionSelected (ClarionAgentActionType.CREATE_THINGS, null);
					}
				}
				
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);
				
				
                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
				
				
            	chosen = CurrentAgent.GetChosenExternalAction(si);
				
				if (chosen != ExternalActionChunk.DO_NOTHING)
				{
	                // Get the selected action
	                String actionLabel = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "type").Value.AsIComparable;
	
					
					Console.WriteLine ("\n\nHas choosen {0} action. Label: {1}\n\n", actionLabel, chosen.LabelAsIComparable.ToString());
					
					
					GoalChunk currentGoal = CurrentAgent.CurrentGoal;
					
					Console.WriteLine ("Current goal: {0}", currentGoal);
	
					
					
					ClarionAgentActionType actionType = (ClarionAgentActionType) Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

	
	                // Call the output event handler
	                if (OnNewExternalActionSelected != null)
	                {
						switch (actionType)
						{
							case ClarionAgentActionType.GO_POSITION:
								
							    Thing targetPosition = null;
							
								tryingToContour = false;
														
								double x = (double) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "x").Value.AsIComparable;
						        double y = (double) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "y").Value.AsIComparable;
								
								Console.WriteLine ("({0}, {1})", x, y);
							

							    if (chosen.LabelAsIComparable.ToString() == "go_deliverySpot")
								{
									if (completedLeaflet != null)
									{
										CurrentAgent.ReceiveFeedback (si, 1.0);
									
										lastMovementFeedback = 1.0;
									
									    targetPosition = deliverySpot;
									
										Console.WriteLine (". Going to delivery spot...");
									}
									else
									{
										CurrentAgent.ReceiveFeedback (si, 0.0);
									
										Console.WriteLine (". Ignoring request to go to delivery spot, since there is no completed leaflet...");
									}
								}
								else if ((chosen.LabelAsIComparable.ToString()).IndexOf ("Jewel") >= 0)
								{		
									if (currentGoal != null ? (currentGoal.LabelAsIComparable.ToString() == FULFILL_LEAFLET) : true)
									{								
										string jewelColor = (string) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "color").Value.AsIComparable;
																
										int index = (int) Enum.Parse(typeof(jewelColors), jewelColor, true);
									
										double needValue = jewelsActivation[index] * 0.5;
										double distance  = 1.0 - (Math.Sqrt (Math.Pow (x - currentCreature.X1, 2) + Math.Pow (y - currentCreature.Y1, 2)) / MAXIMUM_DISTANCE);
									
										double currentFeedbackValue = (needValue + distance) / 2;
										
										Console.WriteLine (". Jewel - lastMovementFeedback: {0}, currentFeedbackValue: {1}", lastMovementFeedback, currentFeedbackValue);
									
									
										if (currentFeedbackValue > lastMovementFeedback)
										{								
											CurrentAgent.ReceiveFeedback (si, currentFeedbackValue);
										
										    lastMovementFeedback = currentFeedbackValue;
										
											String jewelName = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "jewel").Value.AsIComparable;
										
											targetPosition = jewels.Find (item => item.Name == jewelName);										
										}
										else
										{
											// Give a bad feedback...
										
											CurrentAgent.ReceiveFeedback (si, 0.0);
										
										    Console.WriteLine (". Jewel - Bad movement choice. Ignoring it...");
										}
										
									
										Console.WriteLine (". Jewel - needValue: {0}, distance: {1}, feedback: {2}", needValue, distance, currentFeedbackValue);
									}
									else
									{
										// Give bad feedback, since should be going after food...
									
										CurrentAgent.ReceiveFeedback (si, 0.0);
									
									    Console.WriteLine (". Jewel - Bad movement choice; should be going after food. Ignoring it...");
									}
								}
								else if ((chosen.LabelAsIComparable.ToString()).IndexOf ("Food") >= 0)
								{		
									if ((currentGoal != null ? (currentGoal.LabelAsIComparable.ToString() == EAT_FOOD) : false) ||
								    	(jewels.Count() == 0))
									{								
										double distance  = 1.0 - (Math.Sqrt (Math.Pow (x - currentCreature.X1, 2) + Math.Pow (y - currentCreature.Y1, 2)) / MAXIMUM_DISTANCE);
									
										double currentFeedbackValue = distance;
										
										Console.WriteLine (". Food - lastMovementFeedback: {0}, currentFeedbackValue: {1}", lastMovementFeedback, currentFeedbackValue);
									
									
										if (currentFeedbackValue > lastMovementFeedback)
										{								
											CurrentAgent.ReceiveFeedback (si, currentFeedbackValue);
										
										    lastMovementFeedback = currentFeedbackValue;
										
											String foodName = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "food").Value.AsIComparable;
										
											targetPosition = foods.Find (item => item.Name == foodName);										
										}
										else
										{
											// Give a bad feedback...
										
											CurrentAgent.ReceiveFeedback (si, 0.0);
										
										    Console.WriteLine (". Food - Bad movement choice. Ignoring it...");
										}
										
									
										Console.WriteLine (". Food - distance: {0}, feedback: {1}", distance, currentFeedbackValue);
									}
									else
									{
										// Give bad feedback, since should be going after jewels...
									
										CurrentAgent.ReceiveFeedback (si, 0.0);
									
									    Console.WriteLine (". Food - Bad movement choice; should be going after jewels. Ignoring it...");
									}
								}
							
							
								if (targetPosition != null)
								{
									OnNewExternalActionSelected (actionType, targetPosition);
								}
							
								break;
							
							
							case ClarionAgentActionType.COLLECT_JEWEL:
								
								tryingToContour = false;
							
								String jewelName = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "jewel").Value.AsIComparable;
							
								Thing whichJewel = jewels.Find (item => item.Name == jewelName);
							
								if (whichJewel != null)
								{
									OnNewExternalActionSelected (actionType, whichJewel);

									removeCollectedOrEaten (jewelName);
								
									// Reset movement feedback controller
								
									lastMovementFeedback = 0.0;								
								
								
									// Wait for a little bit to avoid WorldServer3D to still "see" the collected jewel...
								
									//Thread.Sleep (200);
								}
								else
								{
									Console.WriteLine ("Target jewel is not found...");
								}
														
							
								break;
							
							
							
							case ClarionAgentActionType.EAT_FOOD:
							
								String foodName = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "food").Value.AsIComparable;
							
								Thing whichFood = foods.Find (item => item.Name == foodName);
							
								tryingToContour = false;
							
								if (whichFood != null)
								{
									OnNewExternalActionSelected (actionType, whichFood);

									removeCollectedOrEaten (foodName);
								
								
									// Reset movement feedback controller
								
									lastMovementFeedback = 0.0;			
								
									// Wait for a little bit to avoid WorldServer3D to still "see" the eaten food...

									Thread.Sleep (200);
								}
								else
								{
									Console.WriteLine ("Target food is not found...");
								}
														
							
								break;
							
							
							
							case ClarionAgentActionType.ROTATE_CLOCKWISE:
							
//								if ((si[InputWallAhead] == CurrentAgent.Parameters.MAX_ACTIVATION) || lastMovementFeedback == 0.0)
							
								tryingToContour = false;
							
								if (lastMovementFeedback == 0.0)
								{
		                    		OnNewExternalActionSelected (actionType, null);
							
									CurrentAgent.ReceiveFeedback (si, 1.0);
								
									lastMovementFeedback = 0.0;
								}
								else
								{
									// Give bad feedback...
								
									CurrentAgent.ReceiveFeedback (si, 0.0);
								
									Console.WriteLine ("No reason to request rotate clockwise...");
								}
														
								break;
							
							
							case ClarionAgentActionType.CONTOUR_OBSTACLE:
							
								if (si[InputWallAhead] == CurrentAgent.Parameters.MAX_ACTIVATION)
								{
									IEnumerable<Thing> whichThings = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= CONTOUR_DISTANCE));
									Thing whichThing = ((whichThings.Count() > 1) || (tryingToContour == true)) ? null : whichThings.First ();
		                    		
									OnNewExternalActionSelected (actionType, whichThing);
							
									//CurrentAgent.ReceiveFeedback (si, 1.0);
								
									lastMovementFeedback = 0.0;
								
									// Let the creature move away from the obstacle
								
									Thread.Sleep (200);
								
									tryingToContour = !tryingToContour;
								}
								else
								{
									// Give bad feedback...
								
									CurrentAgent.ReceiveFeedback (si, 0.0);
								
									Console.WriteLine ("No obstacle to contour...");
								}
							
								break;
							
							
							case ClarionAgentActionType.DELIVER_LEAFLET:						
							
							
								tryingToContour = false;
							
	                    		OnNewExternalActionSelected (actionType, completedLeaflet);

							
								// Now remove the corresponding leaflet jewels from the collectedJewels list
							
								foreach (LeafletItem jewelColor in completedLeaflet.items)
								{
									Console.WriteLine ("Removing {0} {1} jewels from collected", jewelColor.desiredQuantity, jewelColor.color);
								
									for (int i = 0; i < jewelColor.desiredQuantity; i++)
									{
										Thing toRemove = collectedJewels.First (item => item.Material.Color == jewelColor.color);
										collectedJewels.Remove (toRemove);
									}
								}
							
								
								// Zeroes the last feedback...
							
								lastMovementFeedback = 0.0;
								
								break;
						}
	                }
				}
				else
				{
					Console.WriteLine ("Do nothing...");
				}
                			
                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }

        #endregion

        #endregion
    }
}
